/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is Mailnews import code.
 *
 * The Initial Developer of the Original Code is 
 * Håkan Waara <hwaara@gmail.com>.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Postbox, Inc. <contact@postbox-inc.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

#ifdef MOZ_LOGGING
// sorry, this has to be before the pre-compiled header
#define FORCE_PR_LOG /* Allow logging in the release build */
#endif

#include "nsEmlxHelperUtils.h"
#include "nsIFileStreams.h"
#include "nsIBufferedStreams.h"
#include "nsIOutputStream.h"
#include "nsNetUtil.h"
#include "nsCOMPtr.h"
#include "nsObjCExceptions.h"
#include "nsMsgMessageFlags.h"
#include "nsMsgLocalFolderHdrs.h"
#include "msgCore.h"
#include "nsTArray.h"
#include "nsAppleMailImport.h"
#include "prprf.h"
#include "nsMsgUtils.h"

// Use PR_LOG for logging.
#include "prlog.h"
extern PRLogModuleInfo *IMPORTLOGMODULE;  // Logging module

#import <Cocoa/Cocoa.h>


nsresult nsEmlxHelperUtils::ConvertToMozillaStatusFlagsAndDate(const char *aXMLBufferStart, 
                                                        const char *aXMLBufferEnd, 
                                                        PRUint32 *aMozillaStatusFlags,
                                                        nsCString &aOutBuffer)
{
  // create a NSData wrapper around the buffer, so we can use the Cocoa call below
  NSData *metadata = 
    [[[NSData alloc] initWithBytesNoCopy:(void *)aXMLBufferStart length:(aXMLBufferEnd-aXMLBufferStart) freeWhenDone:NO] autorelease];

  // get the XML data as a dictionary
  NSPropertyListFormat format;
  id plist = [NSPropertyListSerialization propertyListFromData:metadata
                                              mutabilityOption:NSPropertyListImmutable
                                                        format:&format
                                              errorDescription:NULL];

  if (!plist)
    return NS_ERROR_FAILURE;

  const PRUint32 dateSent = [[(NSDictionary *)plist objectForKey:@"date-sent"] intValue];
  if (dateSent)
  {
    PRTime dateOfMsg;
    Seconds2PRTime(dateSent, &dateOfMsg);
    char timeBuffer[128];
    PRExplodedTime now;
    PR_ExplodeTime(dateOfMsg, PR_LocalTimeParameters, &now);
    PR_FormatTimeUSEnglish(timeBuffer, sizeof(timeBuffer),
                           "%a %b %d %H:%M:%S %Y",
                           &now);
    aOutBuffer = timeBuffer;
  }
  
  // find the <flags>...</flags> value and convert to int
  const PRUint32 emlxMessageFlags = [[(NSDictionary *)plist objectForKey:@"flags"] intValue];

  
  if (emlxMessageFlags == 0)
    return NS_ERROR_FAILURE;

  if (emlxMessageFlags & nsEmlxHelperUtils::kRead) 
    *aMozillaStatusFlags |= MSG_FLAG_READ;
  if (emlxMessageFlags & nsEmlxHelperUtils::kForwarded)
    *aMozillaStatusFlags |= MSG_FLAG_FORWARDED;
  if (emlxMessageFlags & nsEmlxHelperUtils::kRedirected)
    *aMozillaStatusFlags |= MSG_FLAG_REDIRECTED;
  if (emlxMessageFlags & nsEmlxHelperUtils::kAnswered)
    *aMozillaStatusFlags |= MSG_FLAG_REPLIED;
  if (emlxMessageFlags & nsEmlxHelperUtils::kFlagged)
    *aMozillaStatusFlags |= MSG_FLAG_MARKED;
  if (emlxMessageFlags & nsEmlxHelperUtils::kDeleted)
    *aMozillaStatusFlags |= MSG_FLAG_EXPUNGED;


  return NS_OK;
}

nsresult nsEmlxHelperUtils::GetDateFromHeaders(const char *aMessageBufferStart, const char *aMessageBufferEnd, nsCString &aOutBuffer, PRBool *aHaveDateHeader)
{
  nsTArray<nsCString> allHeaders;
  nsCString headers;
  
  *aHaveDateHeader = PR_FALSE;
  
  const char *bodySeparator = strnstr(aMessageBufferStart, "\n\n", aMessageBufferStart - aMessageBufferEnd);
  if (!bodySeparator)
    return NS_ERROR_UNEXPECTED;
  
  headers.Append(aMessageBufferStart, bodySeparator - aMessageBufferStart);
  ParseString(headers, '\n', allHeaders);
  
  PRBool inReceived;
  nsCString receivedHeaders;
  
  for (PRInt32 i = 0; i < allHeaders.Length(); i++)
  {
    if (strncmp(allHeaders[i].get(), "Date: ", strlen("Date: ")) == 0)
    {
      aOutBuffer = allHeaders[i].get() + strlen("Date: ");
      // Apple Mail to dos have empty date headers.
      if (!aOutBuffer.IsEmpty())
      {
        *aHaveDateHeader = PR_TRUE;
        return NS_OK;
      }
    }
    
    if (strncmp(allHeaders[i].get(), "Received: ", strlen("Received: ")) == 0)
    {
      inReceived = PR_TRUE;
      receivedHeaders += " ";
      receivedHeaders += allHeaders[i].get() + strlen("Received: ");
    }
    else if (inReceived && ((allHeaders[i].get()[0] == '\t') || (allHeaders[i].get()[0] == ' ')))
    {
      receivedHeaders += " ";
      receivedHeaders += allHeaders[i].get();
    }
    else
    {
      inReceived = PR_FALSE;
    }
  }

  if (!receivedHeaders.IsEmpty() && strchr(receivedHeaders.get(), ';'))
  {
    aOutBuffer = strrchr(receivedHeaders.get(), ';') + 1;
  }
           
  return NS_OK;
}

nsresult nsEmlxHelperUtils::FixHeaders(const char *aMessageBufferStart, const char *aMessageBufferEnd, nsCString &aOutBuffer)
{
  nsTArray<nsCString> allHeaders;
  nsCString headers;
  
  const char *bodySeparator = strnstr(aMessageBufferStart, "\n\n", aMessageBufferStart - aMessageBufferEnd);
  if (!bodySeparator)
    return NS_ERROR_UNEXPECTED;
  
  headers.Append(aMessageBufferStart, bodySeparator - aMessageBufferStart);
  ParseString(headers, '\n', allHeaders);

  PRBool fixedHeader = PR_FALSE;
  nsCString dateStr;
  
  for (PRInt32 i = 0; i < allHeaders.Length(); i++)
  {
    if (strncmp(allHeaders[i].get(), "From: ", strlen("From: ")) == 0)
    {
      // look for something like
      // From: G. Gregory Haff, Ph.D. <haffgg@appstate.edu>
      // and fix it.
      if (!strchr(allHeaders[i].get(), '"') && strchr(allHeaders[i].get(), '<') && strchr(allHeaders[i].get(), ','))
      {
        nsCString fromHeader = allHeaders[i];
        fromHeader.Insert('"', fromHeader.Find("<", PR_FALSE, 0, -1) - 1);
        fromHeader.Insert('"', strlen("From: "));
        IMPORT_LOG1("Fixed From header: %s", fromHeader.get());
        allHeaders.ReplaceElementsAt(i, 1, fromHeader);
        fixedHeader = PR_TRUE;
      }
    }
      
    if (strncmp(allHeaders[i].get(), "Date: ", strlen("Date: ")) == 0)
    {
      dateStr = allHeaders[i].get() + strlen("Date: ");
    }
  }
    
  // if we didn't fix any headers, bail out
  if (!fixedHeader)
    return NS_OK;
  
  for (PRInt32 i = 0; i < allHeaders.Length(); i++)
  {
    aOutBuffer += allHeaders[i];
    aOutBuffer += "\n";
  }
  
  aOutBuffer += "\n";
  aOutBuffer.Append(bodySeparator+2, (aMessageBufferEnd-bodySeparator-2));
  return NS_OK;
}


nsresult nsEmlxHelperUtils::ConvertToMboxRD(const char *aMessageBufferStart, const char *aMessageBufferEnd, nsCString &aOutBuffer)
{
  nsTArray<const char *> foundFromLines;

  const char *cur = aMessageBufferStart;
  while (cur < aMessageBufferEnd) {

    const char *foundFromStr = strnstr(cur, "From ", aMessageBufferEnd-cur);

    if (foundFromStr) {
      const char *fromLineStart = foundFromStr;
      if ((fromLineStart == aMessageBufferStart) || (*(fromLineStart-1) == '\n'))
      {
        foundFromLines.AppendElement(fromLineStart);
      }
      
      // advance past the last found From string.
      cur = foundFromStr + 5;

      // look for more From lines.
      continue;
    } 

    break;
  }

  // go through foundFromLines
  if (foundFromLines.Length()) {

    const char *chunkStart = aMessageBufferStart;
    for (unsigned i=0; i<foundFromLines.Length(); ++i) {
      IMPORT_LOG0("Escaped From-line!");
      aOutBuffer.Append(chunkStart, (foundFromLines[i]-chunkStart));
      aOutBuffer.Append(NS_LITERAL_CSTRING(">"));

      chunkStart = foundFromLines[i];
    }
    aOutBuffer.Append(chunkStart, (aMessageBufferEnd-chunkStart));
  }

  return NS_OK;
}

nsresult nsEmlxHelperUtils::AddEmlxMessageToStream(nsILocalFile *aMessage, nsIOutputStream *aOut)
{
  NS_OBJC_BEGIN_TRY_ABORT_BLOCK_NSRESULT;

  // needed to be sure autoreleased objects are released too, which they might not
  // in a C++ environment where the main event loop has no autorelease pool (e.g on a XPCOM thread)
  NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];

  nsresult rv = NS_ERROR_FAILURE;

  nsCAutoString path;
  aMessage->GetNativePath(path);
  IMPORT_LOG1("In AddEmlxMessageToStream and about to import file: %s\n", path.get());

  NSData *data = [NSData dataWithContentsOfFile:[NSString stringWithUTF8String:path.get()]]; 
  if (!data) {
    [pool release];
    return NS_ERROR_FAILURE;
  }

  char *startOfMessageData = NULL;

  // The anatomy of an EMLX file:
  //
  // -------------------------------
  // < A number describing how many bytes ahead there is message data >
  // < Message data >
  // < XML metadata for this message >
  // -------------------------------

  // read the first line of the emlx file, which is a number of how many bytes ahead the actual
  // message data is.
  char *bytes = (char *)[data bytes];
  if (!bytes) {
    [pool release];
    return NS_ERROR_FAILURE;
  }
  
  PRUint64 numberOfBytesToRead = strtol(bytes, &startOfMessageData, 10);
  IMPORT_LOG1("number of bytes to read from the message: %d\n", numberOfBytesToRead);
  // if the number of bytes to read for the message takes us past the end of the data block then bail and abort importing this message
  // as something isn't right here..
  // (char *)[data bytes] + [data length] should point to the very end of our data block. 
  // startOfMessageData + numberOfBytesToRead should point to the start of the xml meta data which should follow the message text.
  if (numberOfBytesToRead <= 0 || !startOfMessageData || startOfMessageData + numberOfBytesToRead > (char *)[data bytes] + [data length]) {
    IMPORT_LOG0("message size appears to be corrupt, abort importing this message.\n");
    [pool release];
    return NS_ERROR_FAILURE;
  }

  // skip whitespace
  while (*startOfMessageData == ' '  || 
         *startOfMessageData == '\n' || 
         *startOfMessageData == '\r' || 
         *startOfMessageData == '\t')
    ++startOfMessageData;
  
  // now read the XML metadata, so we can extract info like which flags (read? replied? flagged? etc) this message has.
  const char *startOfXMLMetadata = startOfMessageData + numberOfBytesToRead;
  const char *endOfXMLMetadata = (char *)[data bytes] + [data length];

  nsCString dateStr;
  PRUint32 x_mozilla_flags = 0;
  ConvertToMozillaStatusFlagsAndDate(startOfXMLMetadata, endOfXMLMetadata, &x_mozilla_flags, dateStr);
  PRBool haveDateHeader = PR_FALSE;
  
  // if we don't have:
  // <key>date-sent</key>
  // <real>1234037968</real>
  // try to get the date from "Date: " or "Received: " headers
  if (dateStr.IsEmpty())
  {
    rv = GetDateFromHeaders(startOfMessageData, (startOfMessageData + numberOfBytesToRead), dateStr, &haveDateHeader);
    if (NS_FAILED(rv)) {
      [pool release];
      return rv;
    }
  }
  
  nsCString fromLine;
  fromLine = "From ";
  fromLine.Append(dateStr);
  fromLine += "\n";

  PRUint32 dummyRv;
  // write the "From " line which is a magic separator in the mbox format
  rv = aOut->Write(fromLine.get(), fromLine.Length(), &dummyRv);
  if (NS_FAILED(rv)) {
    [pool release];
    return rv;
  }

  // write the X-Mozilla-Status header according to which flags we've gathered above.  
  nsCAutoString buf(PR_smprintf(X_MOZILLA_STATUS_FORMAT MSG_LINEBREAK, x_mozilla_flags & 0xFFFF));
  NS_ASSERTION(!buf.IsEmpty(), "printf error with X-Mozilla-Status header");
  if (buf.IsEmpty()) {
    [pool release];
    return rv;
  }

  rv = aOut->Write(buf.get(), buf.Length(), &dummyRv);
  if (NS_FAILED(rv)) {
    [pool release];
    return rv;
  }
  
  // write out X-Mozilla-Keywords header as well to reserve some space for it 
  // in the mbox file.
  rv = aOut->Write(X_MOZILLA_KEYWORDS, X_MOZILLA_KEYWORDS_LEN, &dummyRv);
  if (NS_FAILED(rv)) {
    [pool release];
    return rv;
  }
  
  // write the X-Mozilla-Status-2 header according to which flags we've gathered above.
  char x_mozilla_status_2[40];
  PR_snprintf(x_mozilla_status_2, sizeof(x_mozilla_status_2), X_MOZILLA_STATUS2_FORMAT MSG_LINEBREAK, x_mozilla_flags & 0xFFFF0000);
  rv = aOut->Write(x_mozilla_status_2, strlen(x_mozilla_status_2), &dummyRv);
  if (NS_FAILED(rv)) {
    [pool release];
    return rv;
  }

  // write out the Date header if the original message did not have one
  if (!dateStr.IsEmpty() && !haveDateHeader)
  {
    nsCString dateHeader;
    dateHeader = "Date: ";
    dateHeader.Append(dateStr);
    dateHeader += "\n";
    rv = aOut->Write(dateHeader.get(), dateHeader.Length(), &dummyRv);
    if (NS_FAILED(rv)) {
      [pool release];
      return rv;
    }
  }
  
  // do any conversion needed for the mbox data to be valid mboxrd.
  nsCString convertedData;
  rv = ConvertToMboxRD(startOfMessageData, (startOfMessageData + numberOfBytesToRead), convertedData);
  if (NS_FAILED(rv)) {
    [pool release];
    return rv;
  }
  
  // fix some headers, if necessary
  if (convertedData.IsEmpty())
    rv = FixHeaders(startOfMessageData, (startOfMessageData + numberOfBytesToRead), convertedData);
  else {
    rv = FixHeaders(convertedData.get(), (convertedData.get() + convertedData.Length()), convertedData);
  }
  
  if (NS_FAILED(rv)) {
    [pool release];
    return rv;
  }  
  
  // write the actual message data.
  PRUint32 actualBytesWritten = 0;
  if (convertedData.IsEmpty())
    rv = aOut->Write(startOfMessageData, (PRUint32)numberOfBytesToRead, &actualBytesWritten);
  else {
    rv = aOut->Write(convertedData.get(), convertedData.Length(), &actualBytesWritten);
  }

  if (NS_FAILED(rv)) {
    [pool release];
    return rv;
  }

  IMPORT_LOG1("number of bytes written: %d\n", actualBytesWritten); 

  NS_ASSERTION(actualBytesWritten == (convertedData.IsEmpty() ? numberOfBytesToRead : convertedData.Length()),
               "Didn't write as many bytes as expected for .emlx file?");

  // add newlines to denote the end of this message in the mbox
  NS_NAMED_LITERAL_CSTRING(kEndOfMessage, "\n\n");
  rv = aOut->Write(kEndOfMessage.get(), kEndOfMessage.Length(), &dummyRv);

  [pool release];
  
  IMPORT_LOG1("Exiting AddEmlxMessageToStream for file: %s\n", path.get());
  return rv;

  NS_OBJC_END_TRY_ABORT_BLOCK_NSRESULT;
}
